/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

#include "tests.h"

/* Main test routine */
int main(int argc, char *argv[])
{
	msg_t * acr = NULL;
	msg_avp_t * pi = NULL, *avp1, *avp2;
	
	/* First, initialize the daemon modules */
	INIT_WAAAD();
	
	/* Create the message object from model */
	{
		dict_object_t * acr_model = NULL;

		/* Now find the ACR dictionary object */
		CHECK( 0, dict_search ( DICT_COMMAND, CMD_BY_NAME, "Accounting-Request", &acr_model ) );

		/* Create the instance, using the templates */
		CHECK( 0, msg_new ( acr_model, MSGFL_FROM_TEMPLATE, &acr ) );

		#if 0
		/* For debug: dump the object */
		log_debug("Dumping Accounting-Request message (from template)\n");
		msg_dump_walk(acr);
		#endif
	}
	
	/* Create the Proxy-Info AVP from model */
	{
		dict_object_t * pi_model = NULL;

		/* Now find the ACR dictionary object */
		CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME, "Proxy-Info", &pi_model ) );

		/* Create the instance, using the templates */
		CHECK( 0, msg_avp_new ( pi_model, MSGFL_FROM_TEMPLATE, &pi ) );

		#if 0
		/* For debug: dump the object */
		log_debug("Dumping Proxy-Info AVP (from template)\n");
		msg_dump_walk(pi);
		#endif
		
		/* Check that msg_free does not recursively delete if not told to. */
		CHECK( EINVAL, msg_free( pi, 0 ) );
	}
	
	/* Get a reference to the current last AVP in the message */
	{
		int diff = 0;
		CHECK( 0, msg_browse ( acr, MSG_BRW_LAST_CHILD, &avp1, &diff) );
		CHECK( 1, diff );
		
		/* Check that we cannot add this AVP to another message since it is already linked */
		CHECK( EINVAL, msg_avp_add( pi, MSG_BRW_LAST_CHILD, avp1) );
	}

	/* Now add the Proxy-Info AVP at the end of the message */
	{
		CHECK( 0, msg_avp_add( acr, MSG_BRW_LAST_CHILD, pi) );
		#if 0
		/* For debug: dump the object */
		log_debug("Dumping Accounting-Request with Proxy-Info AVP at the end\n");
		msg_dump_walk(acr);
		#endif
	}
	
	/* Check the last child is now the proxy-Info */
	{
		CHECK( 0, msg_browse ( acr, MSG_BRW_LAST_CHILD, &avp2, NULL) );
		CHECK( pi, avp2 );
	}
	
	/* Check that the avp before the proxy-info is the previous last one */
	{
		int diff = 0;
		CHECK( 0, msg_browse ( pi, MSG_BRW_PREV, &avp2, &diff) );
		CHECK( avp1, avp2 );
		CHECK( 0, diff);
	}
	
	/* Check that there are no AVP after the prox-info */
	CHECK( ENOENT, msg_browse ( pi, MSG_BRW_NEXT, &avp2, NULL) );
	
	/* Test the msg_avp_del function unlinks the object properly */
	{
		dict_object_t * rr_model = NULL;

		/* Now find the ACR dictionary object */
		CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME, "Route-Record", &rr_model ) );

		/* Create the instance, using the templates */
		CHECK( 0, msg_avp_new ( rr_model, MSGFL_FROM_TEMPLATE, &avp1 ) );
		
		/* Add the AVP at the end of the message */
		CHECK( 0, msg_avp_add( pi, MSG_BRW_NEXT, avp1) );
		
		/* Check the last AVP of the message is now this one */
		CHECK( 0, msg_browse ( acr, MSG_BRW_LAST_CHILD, &avp2, NULL) );
		CHECK( avp1, avp2 );
		
		/* Now delete it */
		CHECK( 0, msg_free( avp1, 0 ) );
		
		/* Check the last AVP of the message is back to pi */
		CHECK( 0, msg_browse ( acr, MSG_BRW_LAST_CHILD, &avp2, NULL) );
		CHECK( pi, avp2 );
		
		/* Delete the whole message */
		CHECK( 0, msg_free( acr, 1 ) );
	}
	
	/* Recreate the ACR without any AVP */
	{
		dict_object_t * acr_model = NULL;

		/* Now find the ACR dictionary object */
		CHECK( 0, dict_search ( DICT_COMMAND, CMD_BY_NAME, "Accounting-Request", &acr_model ) );

		/* Create the instance, using the templates */
		CHECK( 0, msg_new ( acr_model, 0, &acr ) );

		#if 0
		/* For debug: dump the object */
		log_debug("Dumping Accounting-Request message (from template)\n");
		msg_dump_walk(acr);
		#endif
		
		/* Check there is no child */
		CHECK( ENOENT, msg_browse ( acr, MSG_BRW_FIRST_CHILD, &avp1, NULL) );
	}
	
	/* Now let's create some additional Dictionary objects for the test */
	{
		/* The constant values used here are totally arbitrary chosen */
		dict_object_t * vendor;
		{
			dict_vendor_data_t vendor_data = { 73565, "Vendor test" };
			CHECK( 0, dict_new ( DICT_VENDOR, &vendor_data , NULL, &vendor ) );
		}
		
		{
			dict_application_data_t app_data = { 73566, "Application test" };
			CHECK( 0, dict_new ( DICT_APPLICATION, &app_data , vendor, NULL ) );
		}
		
		{
			dict_avp_data_t avp_data = { 73567, 0, "AVP Test - no vendor - f32", 0, 0, AVP_TYPE_FLOAT32 };
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , NULL, NULL ) );
		}
		
		{ 
			dict_object_t * type = NULL;
			dict_type_data_t type_data = { AVP_TYPE_INTEGER64, "Int64 test" };
			dict_avp_data_t avp_data = { 73568, 73565, "AVP Test - i64", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_INTEGER64 };
			CHECK( 0, dict_new ( DICT_TYPE, &type_data , NULL, &type ) );
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , type, NULL ) );
		}
		
		{
			dict_object_t * type = NULL;
			dict_type_data_t type_data = { AVP_TYPE_INTEGER32, "Enum32 test" };
			dict_type_enum_data_t val1 = { "i32 const test (val 1)", { .i32 = 1 } };
			dict_type_enum_data_t val2 = { "i32 const test (val 2)", { .i32 = 2 } };
			dict_type_enum_data_t val3 = { "i32 const test (val -5)",{ .i32 = -5 } };
			dict_avp_data_t avp_data = { 73569, 73565, "AVP Test - enumi32", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_INTEGER32 };
			
			CHECK( 0, dict_new ( DICT_TYPE, &type_data , NULL, &type ) );
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val1 , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val2 , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val3 , type, NULL ) );
		}
			
		{ 
			dict_object_t * type = NULL;
			dict_type_data_t type_data = { AVP_TYPE_OCTETSTRING, "OS test" };
			dict_avp_data_t avp_data = { 73570, 73565, "AVP Test - os", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_OCTETSTRING };
			CHECK( 0, dict_new ( DICT_TYPE, &type_data , NULL, &type ) );
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , type, NULL ) );
		}
		
		{
			dict_object_t * type = NULL;
			dict_type_data_t type_data = { AVP_TYPE_OCTETSTRING, "OS enum test" };
			dict_type_enum_data_t val1 = { "os const test (Test)", { .os = { "Test", 4 } } };
			dict_type_enum_data_t val2 = { "os const test (waaad)", { .os = { "waaad", 5 } } };
			dict_type_enum_data_t val3 = { "os const test (waa)", { .os = { "waaad", 3 } } };
			dict_avp_data_t avp_data = { 73571, 73565, "AVP Test - enumos", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_OCTETSTRING };
			
			CHECK( 0, dict_new ( DICT_TYPE, &type_data , NULL, &type ) );
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val1 , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val2 , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val3 , type, NULL ) );
		}
		
		{
			dict_object_t * gavp = NULL;
			dict_avp_data_t avp_data = { 73572, 73565, "AVP Test - grouped", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_GROUPED };
			
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , NULL, &gavp ) );
			
			/* Macro to search AVP and create a rule */		
			#define ADD_RULE( _parent, _vendor, _avpname, _pos, _min, _max, _tpl ) {		\
				dict_object_t * _avp = NULL;							\
				dict_avp_request_t _req = { (_vendor), 0, (_avpname) };				\
				dict_rule_data_t _data;								\
				CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME_AND_VENDOR, &_req, &_avp ) );	\
				_data.rule_avp = _avp;								\
				_data.rule_position = (_pos);							\
				_data.rule_min = (_min);							\
				_data.rule_max = (_max);							\
				_data.rule_template = (_tpl);							\
				CHECK( 0, dict_new ( DICT_RULE, &_data , (_parent), NULL ) );			\
			}
			
			ADD_RULE(gavp, 73565, "AVP Test - os", RULE_OPTIONAL,   -1, -1,  3);
			
		}
			
		{
			dict_object_t * application = NULL;
			dict_object_t * command = NULL;
			dict_cmd_data_t cmd_data = { 73573, "Test-Command-Request", CMD_FLAG_REQUEST, CMD_FLAG_REQUEST };
			
			CHECK( 0, dict_search ( DICT_APPLICATION, APPLICATION_BY_NAME, "Application test", &application ) );
			CHECK( 0, dict_new ( DICT_COMMAND, &cmd_data , application, &command ) );
			ADD_RULE(command, 0,     "AVP Test - no vendor - f32", 	RULE_FIXED_HEAD, -1,  1,  0);
			ADD_RULE(command, 73565, "AVP Test - i64",		RULE_REQUIRED,   -1, -1,  0);
			ADD_RULE(command, 73565, "AVP Test - enumi32", 		RULE_OPTIONAL,   -1, -1,  0);
			ADD_RULE(command, 73565, "AVP Test - os", 		RULE_OPTIONAL,   -1, -1,  0);
			ADD_RULE(command, 73565, "AVP Test - enumos", 		RULE_OPTIONAL,   -1, -1,  0);
			ADD_RULE(command, 73565, "AVP Test - grouped", 		RULE_OPTIONAL,   -1, -1,  0);
		}
	}
	
	/* Now create some values and check the length is correctly handled */
	{
		dict_object_t * cmd_model = NULL;
		msg_t         * msg = NULL;
		dict_object_t * avp_model = NULL;
		msg_avp_t     * avp = NULL;
		avp_value_t	value;
		
		CHECK( 0, dict_search ( DICT_COMMAND, CMD_BY_NAME, "Test-Command-Request", &cmd_model ) );
		
		/* Check an error is trigged if the AVP has no value set */
		{
			CHECK( 0, dict_search ( DICT_AVP,     AVP_BY_NAME,     "AVP Test - no vendor - f32", &avp_model ) );
			
			CHECK( 0, msg_new ( cmd_model, 0, &msg ) );
			CHECK( 0, msg_avp_new ( avp_model, 0, &avp ) );
			
			CHECK( 0, msg_avp_add ( msg, MSG_BRW_FIRST_CHILD, avp ) );
			
			CHECK( EINVAL, msg_update_length ( avp ) );
			CHECK( EINVAL, msg_update_length ( msg ) );
			
			CHECK( 0, msg_free( msg, 1 ) );
		}
		
		/* Check the sizes are handled properly */
		{
			msg_avp_t * avpi = NULL;
			msg_avp_t * avpch = NULL;
			msg_avp_data_t * avpdata = NULL;
			#define ADD_AVP( _parent, _position, _avpi, _avpvendor, _avpname) {			\
				dict_object_t * _avp = NULL;							\
				dict_avp_request_t _req = { (_avpvendor), 0, (_avpname) };			\
				CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME_AND_VENDOR, &_req, &_avp ) );	\
				CHECK( 0, msg_avp_new ( _avp, 0, &_avpi ) );					\
				CHECK( 0, msg_avp_add ( (_parent), (_position), _avpi ) );			\
			}
			/* Create a message with many AVP inside */
			CHECK( 0, msg_new ( cmd_model, 0, &msg ) );
			
			/* Avp no vendor, float32 => size = 12 */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 0,     "AVP Test - no vendor - f32" );
			value.f32 = 3.1415;
			CHECK( 0, msg_avp_setvalue ( avpi, &value ) );
			CHECK( 0, msg_update_length ( avpi ) );
			#if 0
			log_debug("AVP no vendor, value 3.1415:\n");
			msg_dump_one(avpi);
			#endif
			CHECK( 0, msg_avp_data ( avpi, &avpdata ) );
			CHECK( 12, avpdata->avp_len );
			
			/* Add a vendor AVP, integer64 => size = 20 */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - i64" );
			value.i64 = 0x123456789abcdeLL;
			CHECK( 0, msg_avp_setvalue ( avpi, &value ) );
			CHECK( 0, msg_update_length ( avpi ) );
			#if 1
			log_debug("AVP vendor, value 0x123456789abcdeL:\n");
			msg_dump_one(avpi);
			#endif
			CHECK( 0, msg_avp_data ( avpi, &avpdata ) );
			CHECK( 20, avpdata->avp_len );
			
			
			
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - enumi32" );
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - enumi32" );
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - enumi32" );
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - os" );
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - enumos" );
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - enumos" );
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - grouped" );
			  ADD_AVP( avpi, MSG_BRW_LAST_CHILD, avpch, 73565, "AVP Test - os" );
			  ADD_AVP( avpi, MSG_BRW_LAST_CHILD, avpch, 73565, "AVP Test - os" );
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - grouped" );
			  ADD_AVP( avpi, MSG_BRW_LAST_CHILD, avpch, 73565, "AVP Test - os" );
			  ADD_AVP( avpi, MSG_BRW_LAST_CHILD, avpch, 73565, "AVP Test - os" );
			
		  }
				
	
	}
		
		
			

	/* That's all for the tests yet */
	PASSTEST();
} 
	
